from collections.abc import Mapping
from typing import TypeVar
from typing import cast

from ..data.image import ScalarImage
from ..data.subject import Subject
from .transform import Transform

ValueT = TypeVar('ValueT')


class IntensityTransform(Transform):
    """Transform that modifies voxel intensities only."""

    def get_images_dict(self, subject: Subject) -> dict[str, ScalarImage]:
        return subject.get_images_dict(
            intensity_only=True,
            include=self.include,
            exclude=self.exclude,
        )

    def get_images(self, subject: Subject) -> list[ScalarImage]:
        return subject.get_images(
            intensity_only=True,
            include=self.include,
            exclude=self.exclude,
        )

    @staticmethod
    def get_parameter(value: ValueT | Mapping[str, ValueT], name: str) -> ValueT:
        if isinstance(value, Mapping):
            mapping = cast(Mapping[str, ValueT], value)
            return mapping[name]
        return value

    def arguments_are_dict(self) -> bool:
        """Check if main arguments are dict.

        Return `True` if the type of all attributes specified in the
        `args_names` have `dict` type.
        """
        args = list(self._get_named_arguments().values())
        are_dict = [isinstance(arg, dict) for arg in args]
        if all(are_dict):
            return True
        elif not any(are_dict):
            return False
        else:
            message = 'Either all or none of the arguments must be dicts'
            raise ValueError(message)



from __future__ import annotations

from collections.abc import Callable
from collections.abc import Sequence
from pathlib import Path
from typing import TypeAlias

import numpy as np
import torch
from jaxtyping import Shaped

# For typing hints
TypePath: TypeAlias = str | Path
TypeNumber: TypeAlias = int | float
TypeKeys: TypeAlias = Sequence[str] | None
TypeData: TypeAlias = torch.Tensor | np.ndarray
TypeImageTensor: TypeAlias = Shaped[torch.Tensor, 'channels width height depth']  # noqa: F722
TypeImageArray: TypeAlias = Shaped[np.ndarray, 'channels width height depth']  # noqa: F722
TypeImageData: TypeAlias = TypeImageTensor | TypeImageArray
TypeAffineMatrix: TypeAlias = Shaped[np.ndarray, '4 4']  # noqa: F722
TypeDataAffine: TypeAlias = tuple[torch.Tensor, np.ndarray]
TypeImageDataAffine: TypeAlias = tuple[TypeImageTensor, TypeAffineMatrix]
TypeSlice: TypeAlias = int | slice

TypeDoubletInt: TypeAlias = tuple[int, int]
TypeTripletInt: TypeAlias = tuple[int, int, int]
TypeQuartetInt: TypeAlias = tuple[int, int, int, int]
TypeSextetInt: TypeAlias = tuple[int, int, int, int, int, int]

TypeDoubleFloat: TypeAlias = tuple[float, float]
TypeTripletFloat: TypeAlias = tuple[float, float, float]
TypeQuartetFloat: TypeAlias = tuple[float, float, float, float]
TypeSextetFloat: TypeAlias = tuple[float, float, float, float, float, float]

TypeTuple: TypeAlias = int | TypeTripletInt
TypeRangeInt: TypeAlias = int | TypeDoubletInt
TypeSpacing: TypeAlias = float | TypeTripletFloat
TypeSpatialShape: TypeAlias = int | TypeTripletInt
TypeRangeFloat: TypeAlias = float | TypeDoubleFloat
TypeCallable: TypeAlias = Callable[[torch.Tensor], torch.Tensor]
TypeDirection2D: TypeAlias = TypeQuartetFloat
TypeDirection3D: TypeAlias = tuple[
    float,
    float,
    float,
    float,
    float,
    float,
    float,
    float,
    float,
]
TypeDirection: TypeAlias = TypeDirection2D | TypeDirection3D



import enum

import SimpleITK as sitk


class Interpolation(enum.Enum):
    """Interpolation techniques available in ITK.

    For a full quantitative comparison of interpolation methods, you can read
    [Meijering et al. 1999, Quantitative Comparison of Sinc-Approximating Kernels for
    Medical Image Interpolation ](https://link.springer.com/chapter/10.1007/10704282_23).

    Examples:
        >>> import torchio as tio
        >>> transform = tio.RandomAffine(image_interpolation='bspline')
    """

    NEAREST = 'sitkNearestNeighbor'
    """Nearest neighbor interpolation."""

    LINEAR = 'sitkLinear'
    """Linear interpolation."""

    BSPLINE = 'sitkBSpline'
    """B-Spline of order 3 (cubic) interpolation."""

    CUBIC = 'sitkBSpline'
    """Same as `BSPLINE`."""

    GAUSSIAN = 'sitkGaussian'
    """Gaussian interpolation. Sigma is set to 0.8 input pixels and alpha is 4."""

    LABEL_GAUSSIAN = 'sitkLabelGaussian'
    """Smoothly interpolate multi-label images. Sigma is set to 1 input pixel and alpha is 1."""

    HAMMING = 'sitkHammingWindowedSinc'
    """Hamming windowed sinc kernel."""

    COSINE = 'sitkCosineWindowedSinc'
    """Cosine windowed sinc kernel."""

    WELCH = 'sitkWelchWindowedSinc'
    """Welch windowed sinc kernel."""

    LANCZOS = 'sitkLanczosWindowedSinc'
    """Lanczos windowed sinc kernel."""

    BLACKMAN = 'sitkBlackmanWindowedSinc'
    """Blackman windowed sinc kernel."""


def get_sitk_interpolator(interpolation: str) -> int:
    if not isinstance(interpolation, str):
        message = (
            f'Interpolation must be a string, not "{interpolation}"'
            f' of type {type(interpolation)}'
        )
        raise ValueError(message)
    string = getattr(Interpolation, interpolation.upper()).value
    return getattr(sitk, string)



import numpy as np
import torch

from ....data.subject import Subject
from ...fourier import FourierTransform
from ...intensity_transform import IntensityTransform
from .. import RandomTransform


class RandomSpike(RandomTransform, IntensityTransform, FourierTransform):
    r"""Add random MRI spike artifacts.

    Also known as [Herringbone artifact
    ](https://radiopaedia.org/articles/herringbone-artifact),
    crisscross artifact or corduroy artifact, it creates stripes in different
    directions in image space due to spikes in k-space.

    Args:
        num_spikes: Number of spikes $n$ present in k-space.
            If a tuple $(a, b)$ is provided, then
            $n \sim \mathcal{U}(a, b) \cap \mathbb{N}$.
            If only one value $d$ is provided,
            $n \sim \mathcal{U}(0, d) \cap \mathbb{N}$.
            Larger values generate more distorted images.
        intensity: Ratio $r$ between the spike intensity and the maximum
            of the spectrum.
            If a tuple $(a, b)$ is provided, then
            $r \sim \mathcal{U}(a, b)$.
            If only one value $d$ is provided,
            $r \sim \mathcal{U}(-d, d)$.
            Larger values generate more distorted images.
        **kwargs: See [`Transform`][torchio.transforms.Transform] for additional
            keyword arguments.

    Note:
        The execution time of this transform does not depend on the
        number of spikes.
    """

    def __init__(
        self,
        num_spikes: int | tuple[int, int] = (1, 1),
        intensity: float | tuple[float, float] = (1, 3),
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.intensity_range = self._parse_range(
            intensity,
            'intensity_range',
        )
        self.num_spikes_range = self._parse_range(
            num_spikes,
            'num_spikes',
            min_constraint=0,
            type_constraint=int,
        )

    def apply_transform(self, subject: Subject) -> Subject:
        images_dict = self.get_images_dict(subject)
        if not images_dict:
            return subject

        spikes_positions_by_name: dict[str, np.ndarray] = {}
        intensity_by_name: dict[str, float] = {}
        for image_name in images_dict:
            spikes_positions_param, intensity_param = self.get_params(
                self.num_spikes_range,
                self.intensity_range,
            )
            spikes_positions_by_name[image_name] = spikes_positions_param
            intensity_by_name[image_name] = intensity_param
        transform = Spike(
            spikes_positions=spikes_positions_by_name,
            intensity=intensity_by_name,
            **self._get_base_args(),
        )
        transformed = transform(subject)
        assert isinstance(transformed, Subject)
        return transformed

    def get_params(
        self,
        num_spikes_range: tuple[int, int],
        intensity_range: tuple[float, float],
    ) -> tuple[np.ndarray, float]:
        ns_min, ns_max = num_spikes_range
        num_spikes_param = int(torch.randint(ns_min, ns_max + 1, (1,)).item())
        intensity_param = self.sample_uniform(*intensity_range)
        spikes_positions = torch.rand(num_spikes_param, 3).numpy()
        return spikes_positions, intensity_param


class Spike(IntensityTransform, FourierTransform):
    r"""Add MRI spike artifacts.

    Also known as [Herringbone artifact
    ](https://radiopaedia.org/articles/herringbone-artifact),
    crisscross artifact or corduroy artifact, it creates stripes in different
    directions in image space due to spikes in k-space.

    Args:
        spikes_positions:
        intensity: Ratio $r$ between the spike intensity and the maximum
            of the spectrum.
        **kwargs: See [`Transform`][torchio.transforms.Transform] for additional
            keyword arguments.

    Note:
        The execution time of this transform does not depend on the
        number of spikes.
    """

    def __init__(
        self,
        spikes_positions: np.ndarray | dict[str, np.ndarray],
        intensity: float | dict[str, float],
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.spikes_positions = spikes_positions
        self.intensity = intensity
        self.args_names = ['spikes_positions', 'intensity']
        self.invert_transform = False

    def apply_transform(self, subject: Subject) -> Subject:
        for image_name, image in self.get_images_dict(subject).items():
            spikes_positions = self.get_parameter(self.spikes_positions, image_name)
            intensity = self.get_parameter(self.intensity, image_name)
            transformed_tensors = []
            for channel in image.data:
                transformed_tensor = self.add_artifact(
                    channel,
                    np.asarray(spikes_positions),
                    float(intensity),
                )
                transformed_tensors.append(transformed_tensor)
            image.set_data(torch.stack(transformed_tensors))
        return subject

    def add_artifact(
        self,
        tensor: torch.Tensor,
        spikes_positions: np.ndarray,
        intensity_factor: float,
    ):
        if intensity_factor == 0 or len(spikes_positions) == 0:
            return tensor
        spectrum = self.fourier_transform(tensor)
        shape = np.array(spectrum.shape)
        mid_shape = shape // 2
        indices = np.floor(spikes_positions * shape).astype(int)
        for index in indices:
            diff = index - mid_shape
            i, j, k = mid_shape + diff
            artifact = spectrum.cpu().abs().max() * intensity_factor
            if self.invert_transform:
                spectrum[i, j, k] -= artifact
            else:
                spectrum[i, j, k] += artifact
            # If we wanted to add a pure cosine, we should add spikes to both
            # sides of k-space. However, having only one is a better
            # representation og the actual cause of the artifact in real
            # scans. Therefore the next two lines have been removed.
            # #i, j, k = mid_shape - diff
            # #spectrum[i, j, k] = spectrum.max() * intensity_factor
        result = self.inv_fourier_transform(spectrum).real.float()
        return result



from __future__ import annotations

import warnings
from collections.abc import Callable
from collections.abc import Mapping
from types import ModuleType

import numpy as np
import torch

from ..data.image import Image
from ..data.image import ScalarImage
from ..data.subject import Subject
from ..external.imports import get_monai
from .transform import Transform


class MonaiAdapter(Transform):
    """Wraps a MONAI transform for use in TorchIO pipelines.

    This adapter allows using
    [MONAI transforms](https://docs.monai.io/en/stable/transforms.html)
    within TorchIO workflows. Both **dictionary transforms** (e.g.,
    ``NormalizeIntensityd``) and **array transforms** (e.g.,
    ``NormalizeIntensity``) are supported.

    The adapter handles conversion between TorchIO's
    [`Subject`][torchio.Subject] (where values are
    [`Image`][torchio.Image] objects) and MONAI's expected format
    (where values are tensors or
    [`MetaTensor`](https://docs.monai.io/en/stable/data.html#metatensor)
    objects). Image tensors are passed as `MetaTensor` instances with the
    affine matrix embedded, so spatial transforms (e.g., cropping, resizing)
    correctly propagate affine changes.

    **Dictionary transforms** (subclasses of MONAI's ``MapTransform``)
    operate on the full subject dictionary — only the keys specified in
    the MONAI transform are modified.

    **Array transforms** (all other callables) are applied to each
    image in the subject individually, respecting the ``include`` and
    ``exclude`` parameters inherited from
    [`Transform`][torchio.transforms.Transform].

    Args:
        monai_transform: A MONAI transform (dictionary or array) or a
            MONAI-compatible callable. Requires MONAI to be installed,
            e.g., via ``pip install torchio[monai]``.
        **kwargs: See [`Transform`][torchio.transforms.Transform] for
            additional keyword arguments.

    Examples:
        >>> import torch
        >>> import torchio as tio
        >>> from monai.transforms import NormalizeIntensity
        >>> from monai.transforms import NormalizeIntensityd
        >>> from monai.transforms import RandSpatialCropd
        >>> subject = tio.Subject(
        ...     t1=tio.ScalarImage(tensor=torch.randn(1, 64, 64, 64)),
        ...     seg=tio.LabelMap(tensor=torch.ones(1, 64, 64, 64)),
        ... )
        >>> # Array transform — applied to each image
        >>> adapter = tio.MonaiAdapter(NormalizeIntensity())
        >>> transformed = adapter(subject)
        >>> # Dictionary transform — applied to specified keys
        >>> adapter = tio.MonaiAdapter(NormalizeIntensityd(keys=["t1"]))
        >>> transformed = adapter(subject)
        >>> # Spatial dict transform (affine is updated)
        >>> adapter = tio.MonaiAdapter(
        ...     RandSpatialCropd(keys=["t1", "seg"], roi_size=[32, 32, 32]),
        ... )
        >>> transformed = adapter(subject)
        >>> # Inside a Compose pipeline
        >>> pipeline = tio.Compose([
        ...     tio.ToCanonical(),
        ...     tio.MonaiAdapter(NormalizeIntensity()),
        ...     tio.RandomFlip(),
        ... ])
        >>> transformed = pipeline(subject)
    """

    def __init__(
        self,
        monai_transform: Callable,
        **kwargs,
    ):
        super().__init__(**kwargs)
        if not callable(monai_transform):
            message = (
                f'The monai_transform argument must be callable,'
                f' but got {type(monai_transform)}'
            )
            raise TypeError(message)
        self.monai_transform = monai_transform

    def __repr__(self) -> str:
        return f'{self.name}({self.monai_transform})'

    def _add_transform_to_subject_history(self, subject: Subject) -> None:
        """Skip history recording for MONAI adapter transforms.

        The wrapped MONAI transform object cannot be reliably serialized or
        reconstructed by `Subject.get_applied_transforms()`, so the adapter
        is omitted from the subject history. This is similar to container
        transforms such as `Compose` and `OneOf`, which rely on their inner
        transforms being recorded instead.
        """

    def to_hydra_config(self) -> dict[str, object]:
        """Not supported — MONAI transforms are not serializable.

        Raises:
            NotImplementedError: Always.
        """
        message = (
            'MonaiAdapter cannot be exported to a Hydra config because'
            ' the wrapped MONAI transform is not serializable'
        )
        raise NotImplementedError(message)

    def apply_transform(self, subject: Subject) -> Subject:
        """Apply the wrapped MONAI transform to a subject.

        Args:
            subject: TorchIO subject to transform.

        Returns:
            The transformed subject with updated tensor data and, for spatial
            transforms, updated affine matrices.
        """
        monai = get_monai()
        is_dict_transform = isinstance(
            self.monai_transform,
            monai.transforms.MapTransform,
        )
        if is_dict_transform:
            self._apply_dict_transform(subject, monai)
        else:
            self._apply_array_transform(subject, monai)
        return subject

    def _apply_dict_transform(
        self,
        subject: Subject,
        monai: ModuleType,
    ) -> None:
        monai_dict = _subject_to_monai_dict(subject, monai)
        result = self.monai_transform(monai_dict)
        if not isinstance(result, Mapping):
            message = (
                'Expected the MONAI dict transform to return a single'
                f' mapping, but got {type(result)}. Multi-sample'
                ' transforms (returning a list of dicts) are not'
                ' supported by MonaiAdapter.'
            )
            raise TypeError(message)
        _update_subject_from_monai_dict(subject, result, monai)

    def _apply_array_transform(
        self,
        subject: Subject,
        monai: ModuleType,
    ) -> None:
        images_dict = subject.get_images_dict(
            intensity_only=False,
            include=self.include,
            exclude=self.exclude,
        )
        if len(images_dict) > 1 and isinstance(
            self.monai_transform,
            monai.transforms.Randomizable,
        ):
            warnings.warn(
                'Applying a MONAI Randomizable array transform to a'
                ' subject with multiple images. Each image will receive'
                ' different random parameters, which may break spatial'
                ' alignment. Consider using the dictionary version of'
                ' this transform instead (e.g., RandFlipd instead of'
                ' RandFlip).',
                UserWarning,
                stacklevel=4,
            )
        for image in images_dict.values():
            meta_tensor = _image_to_meta_tensor(image, monai)
            result = self.monai_transform(meta_tensor)
            if not isinstance(result, torch.Tensor):
                message = (
                    'Expected a torch.Tensor from the MONAI transform'
                    f' output, but got {type(result)}'
                )
                raise TypeError(message)
            _update_image_from_result(image, result, monai)


def _to_meta_tensor(
    tensor: torch.Tensor,
    affine: np.ndarray,
    monai: ModuleType,
) -> torch.Tensor:
    """Convert a tensor to a MONAI MetaTensor with affine."""
    MetaTensor = monai.data.MetaTensor
    affine_tensor = torch.as_tensor(affine, dtype=torch.float64, device=tensor.device)
    return MetaTensor(tensor, affine=affine_tensor)


def _image_to_meta_tensor(
    image: Image,
    monai: ModuleType,
) -> torch.Tensor:
    """Convert a TorchIO Image to a MONAI MetaTensor."""
    return _to_meta_tensor(image.data, image.affine, monai)


def _unwrap_tensor(result: torch.Tensor, monai: ModuleType) -> torch.Tensor:
    """Extract a plain tensor from a result, unwrapping MetaTensor if needed."""
    MetaTensor = monai.data.MetaTensor
    if isinstance(result, MetaTensor):
        return result.as_tensor()
    return result


def _extract_affine(
    result: torch.Tensor,
    monai: ModuleType,
) -> np.ndarray | None:
    """Extract the affine from a MetaTensor, or return None."""
    MetaTensor = monai.data.MetaTensor
    if isinstance(result, MetaTensor):
        return result.affine.cpu().numpy()
    return None


def _update_image_from_result(
    image: Image,
    result: torch.Tensor,
    monai: ModuleType,
) -> None:
    """Update a TorchIO Image from a MONAI transform result tensor."""
    image.set_data(_unwrap_tensor(result, monai))
    new_affine = _extract_affine(result, monai)
    if new_affine is not None and not np.array_equal(new_affine, image.affine):
        image.affine = new_affine


def _subject_to_monai_dict(
    subject: Subject,
    monai: ModuleType,
) -> dict[str, object]:
    """Convert a Subject to a MONAI-compatible dictionary.

    Image values are converted to MetaTensor instances.
    Non-image values are passed through unchanged.
    """
    monai_dict: dict[str, object] = {}
    for key, value in subject.items():
        if isinstance(value, Image):
            monai_dict[key] = _image_to_meta_tensor(value, monai)
        else:
            monai_dict[key] = value
    return monai_dict


def _update_subject_from_monai_dict(
    subject: Subject,
    monai_dict: Mapping[str, object],
    monai: ModuleType,
) -> None:
    """Update a Subject in-place from the MONAI transform output.

    For keys that correspond to Images in the subject, the tensor data
    is updated. If the output is a MetaTensor with an updated affine,
    the Image affine is also updated.
    """
    for key, value in monai_dict.items():
        if key in subject and isinstance(subject[key], Image):
            image = subject[key]
            assert isinstance(image, Image)
            if not isinstance(value, torch.Tensor):
                message = (
                    'Expected a torch.Tensor from the MONAI transform'
                    f' output, but got {type(value)}'
                )
                raise TypeError(message)
            _update_image_from_result(image, value, monai)
        else:
            # New key or non-image key from MONAI output
            if isinstance(value, torch.Tensor):
                tensor = _unwrap_tensor(value, monai)
                if tensor.ndim == 4:
                    new_affine = _extract_affine(value, monai)
                    if new_affine is None:
                        new_affine = np.eye(4)
                    subject[key] = ScalarImage(
                        tensor=tensor,
                        affine=new_affine,
                    )
                else:
                    subject[key] = tensor
            else:
                subject[key] = value
    subject.update_attributes()



from __future__ import annotations

from collections.abc import Sequence

import numpy as np
import torch

from ....data.subject import Subject
from ....types import TypeTripletInt
from ....types import TypeTuple
from ....utils import to_tuple
from ...intensity_transform import IntensityTransform
from .. import RandomTransform

TypeLocations = Sequence[tuple[TypeTripletInt, TypeTripletInt]]


class RandomSwap(RandomTransform, IntensityTransform):
    r"""Randomly swap patches within an image.

    This is typically used in [context restoration for self-supervised learning
    ](https://www.sciencedirect.com/science/article/pii/S1361841518304699).

    Args:
        patch_size: Tuple of integers $(w, h, d)$ to swap patches
            of size $w \times h \times d$.
            If a single number $n$ is provided, $w = h = d = n$.
        num_iterations: Number of times that two patches will be swapped.
        **kwargs: See [`Transform`][torchio.transforms.Transform] for additional
            keyword arguments.
    """

    def __init__(
        self,
        patch_size: TypeTuple = 15,
        num_iterations: int = 100,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.patch_size = np.array(to_tuple(patch_size))
        self.num_iterations = self._parse_num_iterations(num_iterations)

    @staticmethod
    def _parse_num_iterations(num_iterations):
        if not isinstance(num_iterations, int):
            raise TypeError(
                f'num_iterations must be an int,not {num_iterations}',
            )
        if num_iterations < 0:
            raise ValueError(
                f'num_iterations must be positive,not {num_iterations}',
            )
        return num_iterations

    @staticmethod
    def get_params(
        tensor: torch.Tensor,
        patch_size: np.ndarray,
        num_iterations: int,
    ) -> list[tuple[TypeTripletInt, TypeTripletInt]]:
        si, sj, sk = tensor.shape[-3:]
        spatial_shape = si, sj, sk  # for mypy
        patch_size_values = [int(value) for value in patch_size.tolist()]
        locations = []
        for _ in range(num_iterations):
            first_ini, first_fin = get_random_indices_from_shape(
                spatial_shape,
                patch_size_values,
            )
            while True:
                second_ini, second_fin = get_random_indices_from_shape(
                    spatial_shape,
                    patch_size_values,
                )
                larger_than_initial = np.all(second_ini >= first_ini)
                less_than_final = np.all(second_fin <= first_fin)
                if larger_than_initial and less_than_final:
                    continue  # patches overlap
                else:
                    break  # patches don't overlap
            first_values = [int(value) for value in first_ini.tolist()]
            second_values = [int(value) for value in second_ini.tolist()]
            first_location: TypeTripletInt = (
                first_values[0],
                first_values[1],
                first_values[2],
            )
            second_location: TypeTripletInt = (
                second_values[0],
                second_values[1],
                second_values[2],
            )
            location = first_location, second_location
            locations.append(location)
        return locations

    def apply_transform(self, subject: Subject) -> Subject:
        images_dict = self.get_images_dict(subject)
        if not images_dict:
            return subject

        locations_by_name: dict[str, TypeLocations] = {}
        patch_size_by_name: dict[str, TypeTripletInt] = {}
        broadcast = np.broadcast_to(self.patch_size, 3)
        patch_size = (int(broadcast[0]), int(broadcast[1]), int(broadcast[2]))
        for name, image in images_dict.items():
            locations = self.get_params(
                image.data,
                self.patch_size,
                self.num_iterations,
            )
            locations_by_name[name] = locations
            patch_size_by_name[name] = patch_size
        transform = Swap(
            patch_size=patch_size_by_name,
            locations=locations_by_name,
            **self._get_base_args(),
        )
        transformed = transform(subject)
        assert isinstance(transformed, Subject)
        return transformed


class Swap(IntensityTransform):
    r"""Swap patches within an image.

    This is typically used in [context restoration for self-supervised learning
    ](https://www.sciencedirect.com/science/article/pii/S1361841518304699).

    Args:
        patch_size: Tuple of integers $(w, h, d)$ to swap patches
            of size $w \times h \times d$.
            If a single number $n$ is provided, $w = h = d = n$.
        num_iterations: Number of times that two patches will be swapped.
        **kwargs: See [`Transform`][torchio.transforms.Transform] for additional
            keyword arguments.
    """

    def __init__(
        self,
        patch_size: TypeTripletInt | dict[str, TypeTripletInt],
        locations: TypeLocations | dict[str, TypeLocations],
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.locations = locations
        self.patch_size = patch_size
        self.args_names = ['locations', 'patch_size']
        self.invert_transform = False

    def apply_transform(self, subject: Subject) -> Subject:
        for name, image in self.get_images_dict(subject).items():
            locations = list(self.get_parameter(self.locations, name))
            patch_size = self.get_parameter(self.patch_size, name)
            if self.invert_transform:
                locations.reverse()
            swapped = _swap(image.data, patch_size, locations)
            image.set_data(swapped)
        return subject


def _swap(
    tensor: torch.Tensor,
    patch_size: TypeTuple,
    locations: TypeLocations,
) -> torch.Tensor:
    # Note this function modifies the input in-place
    tensor = tensor.clone()
    patch_size_array = np.array(patch_size)
    for first_ini, second_ini in locations:
        first_ini_array = np.asarray(first_ini)
        second_ini_array = np.asarray(second_ini)
        first_fin = first_ini_array + patch_size_array
        second_fin = second_ini_array + patch_size_array
        first_patch = _crop(tensor, first_ini_array, first_fin)
        second_patch = _crop(tensor, second_ini_array, second_fin).clone()
        _insert(tensor, first_patch, second_ini_array)
        _insert(tensor, second_patch, first_ini_array)
    return tensor


def _insert(
    tensor: torch.Tensor,
    patch: torch.Tensor,
    index_ini: np.ndarray,
) -> None:
    index_fin = index_ini + np.array(patch.shape[-3:])
    i_ini, j_ini, k_ini = index_ini
    i_fin, j_fin, k_fin = index_fin
    tensor[:, i_ini:i_fin, j_ini:j_fin, k_ini:k_fin] = patch


def _crop(
    image: torch.Tensor,
    index_ini: np.ndarray,
    index_fin: np.ndarray,
) -> torch.Tensor:
    i_ini, j_ini, k_ini = index_ini
    i_fin, j_fin, k_fin = index_fin
    return image[:, i_ini:i_fin, j_ini:j_fin, k_ini:k_fin]


def get_random_indices_from_shape(
    spatial_shape: Sequence[int],
    patch_size: Sequence[int],
) -> tuple[np.ndarray, np.ndarray]:
    assert len(spatial_shape) == 3
    assert len(patch_size) in (1, 3)
    shape_array = np.array(spatial_shape)
    patch_size_array = np.array(patch_size)
    max_index_ini_unchecked = shape_array - patch_size_array
    if (max_index_ini_unchecked < 0).any():
        message = (
            f'Patch size {patch_size} cannot be'
            f' larger than image spatial shape {spatial_shape}'
        )
        raise ValueError(message)
    max_index_ini = max_index_ini_unchecked.astype(np.uint16)
    coordinates = []
    for max_coordinate in max_index_ini.tolist():
        if max_coordinate == 0:
            coordinate = 0
        else:
            coordinate = int(torch.randint(max_coordinate, size=(1,)).item())
        coordinates.append(coordinate)
    index_ini = np.array(coordinates, np.uint16)
    index_fin = index_ini + patch_size_array
    return index_ini, index_fin



from collections.abc import Sequence

import numpy as np
import SimpleITK as sitk
import torch

from ....data.io import nib_to_sitk
from ....data.subject import Subject
from ...fourier import FourierTransform
from ...intensity_transform import IntensityTransform
from .. import RandomTransform


class RandomMotion(RandomTransform, IntensityTransform, FourierTransform):
    r"""Add random MRI motion artifact.

    Magnetic resonance images suffer from motion artifacts when the subject
    moves during image acquisition. This transform follows
    [Shaw et al., 2019 ](http://proceedings.mlr.press/v102/shaw19a.html) to
    simulate motion artifacts for data augmentation.

    Args:
        degrees: Tuple $(a, b)$ defining the rotation range in degrees of
            the simulated movements. The rotation angles around each axis are
            $(\theta_1, \theta_2, \theta_3)$,
            where $\theta_i \sim \mathcal{U}(a, b)$.
            If only one value $d$ is provided,
            $\theta_i \sim \mathcal{U}(-d, d)$.
            Larger values generate more distorted images.
        translation: Tuple $(a, b)$ defining the translation in mm of
            the simulated movements. The translations along each axis are
            $(t_1, t_2, t_3)$,
            where $t_i \sim \mathcal{U}(a, b)$.
            If only one value $t$ is provided,
            $t_i \sim \mathcal{U}(-t, t)$.
            Larger values generate more distorted images.
        num_transforms: Number of simulated movements.
            Larger values generate more distorted images.
        image_interpolation: See Interpolation.
        **kwargs: See [`Transform`][torchio.transforms.Transform] for additional
            keyword arguments.

    Warning:
        Large numbers of movements lead to longer execution times for
        3D images.
    """

    def __init__(
        self,
        degrees: float | tuple[float, float] = 10,
        translation: float | tuple[float, float] = 10,  # in mm
        num_transforms: int = 2,
        image_interpolation: str = 'linear',
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.degrees_range = self.parse_degrees(degrees)
        self.translation_range = self.parse_translation(translation)
        if num_transforms < 1 or not isinstance(num_transforms, int):
            message = (
                'Number of transforms must be a strictly positive natural'
                f'number, not {num_transforms}'
            )
            raise ValueError(message)
        self.num_transforms = num_transforms
        self.image_interpolation = self.parse_interpolation(
            image_interpolation,
        )

    def apply_transform(self, subject: Subject) -> Subject:
        images_dict = self.get_images_dict(subject)
        if not images_dict:
            return subject

        times_by_name: dict[str, np.ndarray] = {}
        degrees_by_name: dict[str, np.ndarray] = {}
        translation_by_name: dict[str, np.ndarray] = {}
        interpolation_by_name: dict[str, str] = {}
        for name, image in images_dict.items():
            params = self.get_params(
                self.degrees_range,
                self.translation_range,
                self.num_transforms,
                is_2d=image.is_2d(),
            )
            times_params, degrees_params, translation_params = params
            times_by_name[name] = times_params
            degrees_by_name[name] = degrees_params
            translation_by_name[name] = translation_params
            interpolation_by_name[name] = self.image_interpolation
        transform = Motion(
            degrees=degrees_by_name,
            translation=translation_by_name,
            times=times_by_name,
            image_interpolation=interpolation_by_name,
            **self._get_base_args(),
        )
        transformed = transform(subject)
        assert isinstance(transformed, Subject)
        return transformed

    def get_params(
        self,
        degrees_range: tuple[float, float],
        translation_range: tuple[float, float],
        num_transforms: int,
        perturbation: float = 0.3,
        is_2d: bool = False,
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        # If perturbation is 0, time intervals between movements are constant
        degrees_params = self.get_params_array(
            degrees_range,
            num_transforms,
        )
        translation_params = self.get_params_array(
            translation_range,
            num_transforms,
        )
        if is_2d:  # imagine sagittal (1, A, S)
            degrees_params[:, :-1] = 0  # rotate around Z axis only
            translation_params[:, 2] = 0  # translate in XY plane only
        step = 1 / (num_transforms + 1)
        times = torch.arange(0, 1, step)[1:]
        noise = torch.FloatTensor(num_transforms)
        noise.uniform_(-step * perturbation, step * perturbation)
        times += noise
        times_params = times.numpy()
        return times_params, degrees_params, translation_params

    @staticmethod
    def get_params_array(nums_range: tuple[float, float], num_transforms: int):
        tensor = torch.FloatTensor(num_transforms, 3).uniform_(*nums_range)
        return tensor.numpy()


class Motion(IntensityTransform, FourierTransform):
    r"""Add MRI motion artifact.

    Magnetic resonance images suffer from motion artifacts when the subject
    moves during image acquisition. This transform follows
    [Shaw et al., 2019 ](http://proceedings.mlr.press/v102/shaw19a.html) to
    simulate motion artifacts for data augmentation.

    Args:
        degrees: Sequence of rotations $(\theta_1, \theta_2, \theta_3)$.
        translation: Sequence of translations $(t_1, t_2, t_3)$ in mm.
        times: Sequence of times from 0 to 1 at which the motions happen.
        image_interpolation: See Interpolation.
        **kwargs: See [`Transform`][torchio.transforms.Transform] for additional
            keyword arguments.
    """

    def __init__(
        self,
        degrees: np.ndarray | dict[str, np.ndarray],
        translation: np.ndarray | dict[str, np.ndarray],
        times: np.ndarray | dict[str, np.ndarray],
        image_interpolation: str | dict[str, str],
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.degrees = degrees
        self.translation = translation
        self.times = times
        self.image_interpolation = image_interpolation
        self.args_names = [
            'degrees',
            'translation',
            'times',
            'image_interpolation',
        ]

    def apply_transform(self, subject: Subject) -> Subject:
        for image_name, image in self.get_images_dict(subject).items():
            degrees = self.get_parameter(self.degrees, image_name)
            translation = self.get_parameter(self.translation, image_name)
            times = self.get_parameter(self.times, image_name)
            image_interpolation = self.get_parameter(
                self.image_interpolation, image_name
            )
            result_arrays = []
            for channel in image.data:
                sitk_image = nib_to_sitk(
                    channel[np.newaxis],
                    image.affine,
                    force_3d=True,
                )
                transforms = self.get_rigid_transforms(
                    degrees,
                    translation,
                    sitk_image,
                )
                transformed_channel = self.add_artifact(
                    sitk_image,
                    transforms,
                    times,
                    image_interpolation,
                )
                result_arrays.append(transformed_channel)
            result = np.stack(result_arrays)
            image.set_data(torch.as_tensor(result))
        return subject

    def get_rigid_transforms(
        self,
        degrees_params: np.ndarray,
        translation_params: np.ndarray,
        image: sitk.Image,
    ) -> list[sitk.Euler3DTransform]:
        center_ijk = np.array(image.GetSize()) / 2
        center_lps = image.TransformContinuousIndexToPhysicalPoint(center_ijk)
        identity = np.eye(4)
        matrices = [identity]
        zipped = zip(degrees_params, translation_params, strict=True)
        for degrees, translation in zipped:
            radians = np.radians(degrees).tolist()
            motion = sitk.Euler3DTransform()
            motion.SetCenter(center_lps)
            motion.SetRotation(*radians)
            motion.SetTranslation(translation.tolist())
            motion_matrix = self.transform_to_matrix(motion)
            matrices.append(motion_matrix)
        transforms = [self.matrix_to_transform(m) for m in matrices]
        return transforms

    @staticmethod
    def transform_to_matrix(transform: sitk.Euler3DTransform) -> np.ndarray:
        matrix = np.eye(4)
        rotation = np.array(transform.GetMatrix()).reshape(3, 3)
        matrix[:3, :3] = rotation
        matrix[:3, 3] = transform.GetTranslation()
        return matrix

    @staticmethod
    def matrix_to_transform(matrix: np.ndarray) -> sitk.Euler3DTransform:
        transform = sitk.Euler3DTransform()
        rotation = matrix[:3, :3].flatten().tolist()
        transform.SetMatrix(rotation)
        transform.SetTranslation(matrix[:3, 3])
        return transform

    def resample_images(
        self,
        image: sitk.Image,
        transforms: Sequence[sitk.Euler3DTransform],
        interpolation: str,
    ) -> list[sitk.Image]:
        floating = reference = image
        default_value = np.float64(sitk.GetArrayViewFromImage(image).min())
        transforms = transforms[1:]  # first is identity
        images = [image]  # first is identity
        for transform in transforms:
            interpolator = self.get_sitk_interpolator(interpolation)
            resampler = sitk.ResampleImageFilter()
            resampler.SetInterpolator(interpolator)
            resampler.SetReferenceImage(reference)
            resampler.SetOutputPixelType(sitk.sitkFloat32)
            resampler.SetDefaultPixelValue(default_value)
            resampler.SetTransform(transform)
            resampled = resampler.Execute(floating)
            images.append(resampled)
        return images

    @staticmethod
    def sort_spectra(spectra: list[torch.Tensor], times: np.ndarray):
        """Use original spectrum to fill the center of k-space."""
        num_spectra = len(spectra)
        if np.any(times > 0.5):
            index = np.where(times > 0.5)[0].min()
        else:
            index = num_spectra - 1
        spectra[0], spectra[index] = spectra[index], spectra[0]

    def add_artifact(
        self,
        image: sitk.Image,
        transforms: Sequence[sitk.Euler3DTransform],
        times: np.ndarray,
        interpolation: str,
    ):
        images = self.resample_images(image, transforms, interpolation)
        spectra = []
        for image in images:
            array = sitk.GetArrayFromImage(image).transpose()  # sitk to np
            spectrum = self.fourier_transform(torch.from_numpy(array))
            spectra.append(spectrum)
        self.sort_spectra(spectra, times)
        result_spectrum = torch.empty_like(spectra[0])
        last_index = result_spectrum.shape[2]
        indices_array = (last_index * times).astype(int)
        indices = [int(value) for value in indices_array.tolist()]
        indices.append(last_index)
        ini = 0
        for spectrum, fin in zip(spectra, indices, strict=True):
            result_spectrum[..., ini:fin] = spectrum[..., ini:fin]
            ini = fin
        result_image = self.inv_fourier_transform(result_spectrum).real.float()
        return result_image



"""Most of this code is from torchvision.

I will remove all this once verbosity is reduced. More info:
https://github.com/pytorch/vision/issues/2830
"""

from __future__ import annotations

import gzip
import hashlib
import os
import tarfile
import zipfile
from urllib import error
from urllib import request

from torch.hub import tqdm

from .types import TypePath


def calculate_md5(fpath, chunk_size=1024 * 1024):
    md5 = hashlib.md5()
    with open(fpath, 'rb') as f:
        for chunk in iter(lambda: f.read(chunk_size), b''):
            md5.update(chunk)
    return md5.hexdigest()


def check_md5(fpath, md5, **kwargs):
    return md5 == calculate_md5(fpath, **kwargs)


def check_integrity(fpath, md5=None):
    if not os.path.isfile(fpath):
        return False
    if md5 is None:
        return True
    return check_md5(fpath, md5)


def gen_bar_updater():
    pbar = tqdm(total=None)

    def bar_update(count, block_size, total_size):
        if pbar.total is None and total_size:
            pbar.total = total_size
        progress_bytes = count * block_size
        pbar.update(progress_bytes - pbar.n)

    return bar_update


# Adapted from torchvision, removing print statements
def download_and_extract_archive(
    url: str,
    download_root: TypePath,
    extract_root: TypePath | None = None,
    filename: TypePath | None = None,
    md5: str | None = None,
    remove_finished: bool = False,
) -> None:
    download_root = os.path.expanduser(download_root)
    if extract_root is None:
        extract_root = download_root
    if not filename:
        filename = os.path.basename(url)
    download_url(url, download_root, filename, md5)
    archive = os.path.join(download_root, filename)
    extract_archive(archive, extract_root, remove_finished)


def _is_tarxz(filename):
    return filename.endswith('.tar.xz')


def _is_tar(filename):
    return filename.endswith('.tar')


def _is_targz(filename):
    return filename.endswith('.tar.gz')


def _is_tgz(filename):
    return filename.endswith('.tgz')


def _is_gzip(filename):
    return filename.endswith('.gz') and not filename.endswith('.tar.gz')


def _is_zip(filename):
    return filename.endswith('.zip')


def extract_archive(from_path, to_path=None, remove_finished=False):
    if to_path is None:
        to_path = os.path.dirname(from_path)

    if _is_tar(from_path):
        with tarfile.open(from_path, 'r') as tar:
            tar.extractall(path=to_path)
    elif _is_targz(from_path) or _is_tgz(from_path):
        with tarfile.open(from_path, 'r:gz') as tar:
            tar.extractall(path=to_path)
    elif _is_tarxz(from_path):
        with tarfile.open(from_path, 'r:xz') as tar:
            tar.extractall(path=to_path)
    elif _is_gzip(from_path):
        stem = os.path.splitext(os.path.basename(from_path))[0]
        to_path = os.path.join(to_path, stem)
        with open(to_path, 'wb') as out_f, gzip.GzipFile(from_path) as zip_f:
            out_f.write(zip_f.read())
    elif _is_zip(from_path):
        with zipfile.ZipFile(from_path, 'r') as z:
            z.extractall(to_path)
    else:
        raise ValueError(f'Extraction of {from_path} not supported')

    if remove_finished:
        os.remove(from_path)


# Adapted from torchvision, removing print statements
def download_url(
    url: str,
    root: TypePath,
    filename: TypePath | None = None,
    md5: str | None = None,
) -> None:
    """Download a file from a url and place it in root.

    Args:
        url: URL to download file from
        root: Directory to place downloaded file in
        filename: Name to save the file under.
            If `None`, use the basename of the URL
        md5: MD5 checksum of the download. If None, do not check
    """

    root = os.path.expanduser(root)
    if not filename:
        filename = os.path.basename(url)
    fpath = os.path.join(root, filename)
    os.makedirs(root, exist_ok=True)
    # check if file is already present locally
    if not check_integrity(fpath, md5):
        try:
            print('Downloading ' + url + ' to ' + fpath)  # noqa: T201
            request.urlretrieve(
                url,
                fpath,
                reporthook=gen_bar_updater(),
            )
        except (error.URLError, OSError) as e:
            if url[:5] == 'https':
                url = url.replace('https:', 'http:')
                message = (
                    'Failed download. Trying https -> http instead. Downloading '
                    + url
                    + ' to '
                    + fpath
                )
                print(message)  # noqa: T201
                request.urlretrieve(
                    url,
                    fpath,
                    reporthook=gen_bar_updater(),
                )
            else:
                raise e
        # check integrity of downloaded file
        if not check_integrity(fpath, md5):
            raise RuntimeError('File not found or corrupted.')



import numpy as np
import scipy.ndimage as ndi
import torch

from ....data.subject import Subject
from ....types import TypeData
from ....types import TypeSextetFloat
from ....types import TypeTripletFloat
from ...intensity_transform import IntensityTransform
from .. import RandomTransform


class RandomBlur(RandomTransform, IntensityTransform):
    r"""Blur an image using a random-sized Gaussian filter.

    Args:
        std: Tuple $(a_1, b_1, a_2, b_2, a_3, b_3)$ representing the
            ranges (in mm) of the standard deviations
            $(\sigma_1, \sigma_2, \sigma_3)$ of the Gaussian kernels used
            to blur the image along each axis, where
            $\sigma_i \sim \mathcal{U}(a_i, b_i)$.
            If two values $(a, b)$ are provided,
            then $\sigma_i \sim \mathcal{U}(a, b)$.
            If only one value $x$ is provided,
            then $\sigma_i \sim \mathcal{U}(0, x)$.
            If three values $(x_1, x_2, x_3)$ are provided,
            then $\sigma_i \sim \mathcal{U}(0, x_i)$.
        **kwargs: See [`Transform`][torchio.transforms.Transform] for additional
            keyword arguments.
    """

    def __init__(self, std: float | tuple[float, float] = (0, 2), **kwargs):
        super().__init__(**kwargs)
        self.std_ranges = self.parse_params(std, None, 'std', min_constraint=0)

    def apply_transform(self, subject: Subject) -> Subject:
        images_dict = self.get_images_dict(subject)
        if not images_dict:
            return subject

        std_by_name: dict[str, TypeTripletFloat] = {}
        for name in images_dict:
            std_by_name[name] = self.get_params(self.std_ranges)
        transform = Blur(std=std_by_name, **self._get_base_args())
        transformed = transform(subject)
        assert isinstance(transformed, Subject)
        return transformed

    def get_params(self, std_ranges: TypeSextetFloat) -> TypeTripletFloat:
        sx, sy, sz = self.sample_uniform_sextet(std_ranges)
        return sx, sy, sz


class Blur(IntensityTransform):
    r"""Blur an image using a Gaussian filter.

    Args:
        std: Tuple $(\sigma_1, \sigma_2, \sigma_3)$ representing the
            the standard deviations (in mm) of the Gaussian kernels used to
            blur the image along each axis.
        **kwargs: See [`Transform`][torchio.transforms.Transform] for additional
            keyword arguments.
    """

    def __init__(
        self,
        std: TypeTripletFloat | dict[str, TypeTripletFloat],
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.std = std
        self.args_names = ['std']

    def apply_transform(self, subject: Subject) -> Subject:
        for name, image in self.get_images_dict(subject).items():
            stds = self.get_parameter(self.std, name)
            repets = image.num_channels, 1
            stds_channels: np.ndarray = np.tile(stds, repets)
            transformed_tensors = []
            for std, channel in zip(stds_channels, image.data, strict=True):
                transformed_tensor = blur(
                    channel,
                    image.spacing,
                    std,
                )
                transformed_tensors.append(transformed_tensor)
            image.set_data(torch.stack(transformed_tensors))
        return subject


def blur(
    data: TypeData,
    spacing: TypeTripletFloat,
    std_physical: TypeTripletFloat,
) -> torch.Tensor:
    assert data.ndim == 3
    # For example, if the standard deviation of the kernel is 2 mm and the
    # image spacing is 0.5 mm/voxel, the kernel should be
    # (2 mm / 0.5 mm/voxel) = 4 voxels wide
    std_voxel = np.array(std_physical) / np.array(spacing)
    blurred = ndi.gaussian_filter(data, std_voxel)
    tensor = torch.as_tensor(blurred)
    return tensor



from collections.abc import Sequence

from ...transform import TypeMaskingMethod
from .remap_labels import RemapLabels


class RemoveLabels(RemapLabels):
    r"""Remove labels from a label map.

    The removed labels are remapped to the background label.

    This transformation is not [invertible ](../index.md#invertibility).

    Args:
        labels: A sequence of label integers that will be removed.
        background_label: integer that specifies which label is considered to
            be background (typically, `0`).
        masking_method: See [`RemapLabels`][torchio.transforms.RemapLabels].
        **kwargs: See [`Transform`][torchio.transforms.Transform] for additional
            keyword arguments.

    """

    def __init__(
        self,
        labels: Sequence[int],
        background_label: int = 0,
        masking_method: TypeMaskingMethod = None,
        **kwargs,
    ):
        remapping = {label: background_label for label in labels}
        super().__init__(remapping, masking_method, **kwargs)
        self.labels = labels
        self.background_label = background_label
        self.masking_method = masking_method
        self.args_names = ['labels', 'background_label', 'masking_method']

    def is_invertible(self):
        return False



import torch

from ....data.subject import Subject
from ...intensity_transform import IntensityTransform
from .. import RandomTransform


class RandomNoise(RandomTransform, IntensityTransform):
    r"""Add Gaussian noise with random parameters.

    Add noise sampled from a normal distribution with random parameters.

    Args:
        mean: Mean $\mu$ of the Gaussian distribution
            from which the noise is sampled.
            If two values $(a, b)$ are provided,
            then $\mu \sim \mathcal{U}(a, b)$.
            If only one value $d$ is provided,
            $\mu \sim \mathcal{U}(-d, d)$.
        std: Standard deviation $\sigma$ of the Gaussian distribution
            from which the noise is sampled.
            If two values $(a, b)$ are provided,
            then $\sigma \sim \mathcal{U}(a, b)$.
            If only one value $d$ is provided,
            $\sigma \sim \mathcal{U}(0, d)$.
        **kwargs: See [`Transform`][torchio.transforms.Transform] for additional
            keyword arguments.
    """

    def __init__(
        self,
        mean: float | tuple[float, float] = 0,
        std: float | tuple[float, float] = (0, 0.25),
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.mean_range = self._parse_range(mean, 'mean')
        self.std_range = self._parse_range(std, 'std', min_constraint=0)

    def apply_transform(self, subject: Subject) -> Subject:
        images_dict = self.get_images_dict(subject)
        if not images_dict:
            return subject

        means_by_name: dict[str, float] = {}
        stds_by_name: dict[str, float] = {}
        seeds_by_name: dict[str, int] = {}
        for image_name in images_dict:
            mean, std, seed = self.get_params(self.mean_range, self.std_range)
            means_by_name[image_name] = mean
            stds_by_name[image_name] = std
            seeds_by_name[image_name] = seed
        transform = Noise(
            mean=means_by_name,
            std=stds_by_name,
            seed=seeds_by_name,
            **self._get_base_args(),
        )
        transformed = transform(subject)
        assert isinstance(transformed, Subject)
        return transformed

    def get_params(
        self,
        mean_range: tuple[float, float],
        std_range: tuple[float, float],
    ) -> tuple[float, float, int]:
        mean = self.sample_uniform(*mean_range)
        std = self.sample_uniform(*std_range)
        seed = self._get_random_seed()
        return mean, std, seed


class Noise(IntensityTransform):
    r"""Add Gaussian noise.

    Add noise sampled from a normal distribution.

    Args:
        mean: Mean $\mu$ of the Gaussian distribution
            from which the noise is sampled.
        std: Standard deviation $\sigma$ of the Gaussian distribution
            from which the noise is sampled.
        seed: Seed for the random number generator.
        **kwargs: See [`Transform`][torchio.transforms.Transform] for additional
            keyword arguments.
    """

    def __init__(
        self,
        mean: float | dict[str, float],
        std: float | dict[str, float],
        seed: int | dict[str, int],
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.noise_mean: float | dict[str, float] = mean
        self.noise_std: float | dict[str, float] = std
        self.seed: int | dict[str, int] = seed
        self.invert_transform = False
        self.args_names = ['mean', 'std', 'seed']

    def apply_transform(self, subject: Subject) -> Subject:
        for name, image in self.get_images_dict(subject).items():
            mean = self.get_parameter(self.noise_mean, name)
            std = self.get_parameter(self.noise_std, name)
            seed = self.get_parameter(self.seed, name)
            with self._use_seed(seed):
                noise = get_noise(image.data, mean, std)
            if self.invert_transform:
                noise *= -1
            image.set_data(image.data + noise)
        return subject

    def _get_named_arguments(self) -> dict[str, object]:
        return {
            'mean': self.noise_mean,
            'std': self.noise_std,
            'seed': self.seed,
        }


def get_noise(tensor: torch.Tensor, mean: float, std: float) -> torch.Tensor:
    return torch.randn(*tensor.shape) * std + mean
